
/*************************************************************************
 *
 * All portions are copyrighted by their respective authors.
 *
 * This file is part of IVMan (ivm).
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Troll Tech AS of Norway and appearing in the file
 * LICENSE.QPL included in the packaging of this file.
 * 
 * See http://www.troll.no/qpl for QPL licensing information.
 *
 * $Id: manager.c,v 1.64 2005/11/22 00:25:58 ro_han Exp $
 ************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <dirent.h>
#include <mntent.h>
#include <errno.h>

#include <locale.h>

#include <linux/cdrom.h>
#include <linux/sysctl.h>

#include <glib.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <libhal.h>


#include "common.h"
#include "hal_interface.h"
#include "hal_compat.h"
#include "manager.h"
#include "debug.h"
#include "IvmConfig/IvmConfigProperties.h"
#include "IvmConfig/IvmConfigActions.h"
#include "hal_psi.h"
#include "log.h"
#include "daemonize.h"

#include "userconfigs.h"

gchar *homeDir;

/// Command to be used for mounting, as passed to ivm_run_command() .
char *mount_command;

/// Command to be used for unmounting, as passed to ivm_run_command() .
char *umount_command;

/// Command-line option: run in 'root mode' when not root.
gboolean systemOpt = FALSE;

/// Command-line option: don't fork.
gboolean noForkOpt = FALSE;

/// Command-line option: output debugging info.
gboolean debugOpt = FALSE;


/**
 * Find the commands used to mount and unmount volumes on this system.
 */
void set_mount_command()
{
    mount_command = NULL;
    umount_command = NULL;
    // If command used to mount is preset, just return it
    if ((cfg_base->mountcommand != NULL)
        && (cfg_base->umountcommand != NULL))
    {
        mount_command = cfg_base->mountcommand;
        umount_command = cfg_base->umountcommand;
        return;
    }

    if ((cfg_base->mountcommand != NULL)
        || (cfg_base->umountcommand != NULL))
    {
        DEBUG(_
              ("Only one of 'mountcommand', 'umountcommand' was specified "
               "in IvmConfigBase.xml.  Both must be specified for the "
               "options to take effect.\n"));
    }

    DEBUG(_
          ("No mount command was specified in IvmConfigBase.xml.  Ivman "
           "will try to automatically detect the command to use. "
           "If Ivman incorrectly detects the program(s) available on your "
           "system, first make sure the program(s) are in the default "
           "shell PATH, then please report it as a bug.\n"));

    // OK, now try to find a mount program.
    // mount programs in order of preference:
    // - pmount-hal (for user-mode Ivman only)
    // - pmount
    // - mount
    gchar *argv[4];

    argv[0] = "/bin/sh";
    argv[1] = "-c";
    argv[3] = 0;
    gint exitStatus;
    GError *error;
    gboolean success;


    // Test if pmount-hal exists
    if (usermode)
    {
        argv[2] = "pmount-hal &> /dev/null";
        success = g_spawn_sync(NULL,
                               argv, NULL, 0,
                               NULL, NULL,
                               NULL, NULL, &exitStatus, &error);

        if (!success)
        {
            DEBUG(_
                  ("Error while trying to detect pmount-hal: %s\n"),
                  error->message);
            g_error_free(error);
            return;
        }

        if (WEXITSTATUS(exitStatus) == 127)
        {
            DEBUG(_("pmount-hal was not found on your system.\n"));
        }
        else
        {
            argv[2] = "pumount &> /dev/null";
            success = g_spawn_sync(NULL, argv, NULL, 0, NULL,
                                   NULL, NULL, NULL, &exitStatus, &error);

            if (!success)
            {
                DEBUG(_
                      ("Error while trying to detect pumount: %s\n"),
                      error->message);
                g_error_free(error);
                return;
            }

            if (WEXITSTATUS(exitStatus) == 127)
            {
                DEBUG(_
                      ("pmount-hal was found on your system, but pumount "
                       "wasn't!\n"));
            }
            else
            {
                DEBUG(_
                      ("pmount-hal was found on your system. It will be "
                       "used for mounting.\n"));
                mount_command = "pmount-hal '$hal.info.udi$'";
                umount_command = "pumount '$hal.block.device$'";
                return;
            }
        }
    }
    else
    {
        DEBUG(_
              ("pmount-hal detection skipped, as we are a root instance "
               "of Ivman.  pmount-hal is only used for user instances.\n"));
    }


    // Now try looking for regular pmount.
    argv[2] = "pmount &> /dev/null";
    success =
        g_spawn_sync(NULL, argv,
                     NULL, 0, NULL, NULL, NULL, NULL, &exitStatus, &error);

    if (!success)
    {

        DEBUG(_("Error while trying to detect pmount: %s\n"),
              error->message);

        g_error_free(error);
        return;
    }

    if (WEXITSTATUS(exitStatus) == 127)
    {
        DEBUG(_("pmount was not found on your system.\n"));
    }
    else
    {
        argv[2] = "pumount &> /dev/null";
        success = g_spawn_sync(NULL, argv, NULL, 0, NULL,
                               NULL, NULL, NULL, &exitStatus, &error);

        if (!success)
        {
            DEBUG(_
                  ("Error while trying to detect pumount: %s\n"),
                  error->message);
            g_error_free(error);
            return;
        }

        if (WEXITSTATUS(exitStatus) == 127)
        {
            DEBUG(_
                  ("pmount was found on your system, but pumount "
                   "wasn't!\n"));
        }

        else
        {
            // If root, make sure -- u exists.
            if (usermode)
            {
                DEBUG(_
                      ("pmount was found on your system. It will be used "
                       "for mounting.\n"));
                mount_command =
                    "pmount '$hal.block.device$' "
                    "'$hal.volume.policy.desired_mount_point$'";
                umount_command = "pumount '$hal.block.device$'";
				return;
            }


            GString *pmount_base_command = g_string_new("pmount");

            pmount_base_command =
                g_string_append(pmount_base_command, " --help | ");

            GString *pmount_umask_check =
                g_string_new(pmount_base_command->str);

            pmount_umask_check =
                g_string_append(pmount_umask_check, GREP_PATH
                                " -- '-u <' &> /dev/null");

            argv[2] = pmount_umask_check->str;
            success =
                g_spawn_sync(NULL, argv, NULL, 0,
                             NULL, NULL, NULL, NULL, &exitStatus, &error);

            g_string_free(pmount_umask_check, TRUE);

            if (!success)
            {

                DEBUG(_
                      ("Error while trying to execute pmount: %s\n"),
                      error->message);
                g_error_free(error);
                g_string_free(pmount_base_command, TRUE);
                return;
            }

            if (WIFEXITED(exitStatus) && (WEXITSTATUS(exitStatus) == 0))
            {

                DEBUG(_("pmount accepts -u <umask>\n"));
                DEBUG(_
                      ("pmount was found on your system. It will be used "
                       "for mounting.\n"));
                GString *command = g_string_new("pmount -u ");

                g_string_append(command, cfg_base->umask);
                g_string_append(command, " '$hal.block.device$' "
                                "'$hal.volume.policy.desired_mount_point$'");
                char *final_command = command->str;

                g_string_free(command, FALSE);

                mount_command = final_command;
                umount_command = "pumount '$hal.block.device$'";
                return;
            }



            DEBUG(_("pmount does not accept -u <umask>"));

            DEBUG(_
                  ("WARNING: pmount does not accept -u <umask>, so I will "
                   "not mount volumes with it (since they would only be "
                   "accessible by the %s user account, which is probably "
                   "not what you want). Upgrade pmount.\n"),
                  cfg_base->user);
        }
    }


    // At the bottom of the barrel is plain old mount ...
    argv[2] = "mount &> /dev/null";

    success =
        g_spawn_sync(NULL, argv, NULL, 0, NULL, NULL,
                     NULL, NULL, &exitStatus, &error);

    if (!success)
    {
        DEBUG(_("Error while trying to detect mount: %s\n"),
              error->message);
        g_error_free(error);
        return;
    }


    if (WEXITSTATUS(exitStatus) == 127)
    {
        DEBUG(_("mount was not found on your system.\n"));
    }

    else
    {
        argv[2] = "umount &> /dev/null";

        success =
            g_spawn_sync(NULL, argv, NULL, 0, NULL, NULL,
                         NULL, NULL, &exitStatus, &error);

        if (!success)
        {
            DEBUG(_
                  ("Error while trying to detect umount: %s\n"),
                  error->message);
            g_error_free(error);
            return;
        }


        if (WEXITSTATUS(exitStatus) == 127)
        {
            DEBUG(_
                  ("mount was found on your system, but umount was "
                   "not!\n"));
        }
        else
        {
            DEBUG(_
                  ("mount was found on your system. It will be used for "
                   "mounting.  This means you need fstab entries for the "
                   "devices you want mounted, or you need to use "
                   "fstab-sync. It is highly recommended you get pmount "
                   "instead.\n"));

            mount_command = "mount '$hal.block.device$'";
            umount_command = "umount '$hal.block.device$'";
			return;
        }
    }


    // Couldn't find anything !!!
    return;

}



// when passed a filename of a configuration file, returns the full
// path which should be used to load that file
char *ivm_get_config_file(char const *const filename)
{

    GString *configFile;

    if (usermode == TRUE)
    {
        configFile = g_string_new((char *) strdup(homeDir));
        configFile = g_string_append(configFile, "/.ivman/");
        configFile = g_string_append(configFile, filename);
    }

    else
    {

        configFile = g_string_new(SYSCONFDIR);
        configFile = g_string_append(configFile, "/ivman/");
        configFile = g_string_append(configFile, filename);
    }

    char *ret = configFile->str;

    g_string_free(configFile, FALSE);
    return ret;

}



// when running in user mode, tests for config files, and creates them
// if they do not already exist.
void ivm_test_configs()
{
    FILE *file;
    GString *filename;
    char *content;

    GString *directory = g_string_new((char *) strdup(homeDir));

    directory = g_string_append(directory, "/.ivman/");

    DIR *dir = opendir(directory->str);

    if (dir == NULL)
    {
        DEBUG(_("Settings directory does not exist, attempting "
                "to create it...\n"));

        // make the necessary directory if it does not exist
        GString *command = g_string_new(MKDIR_PATH);

        command = g_string_append(command, " -p ");
        command = g_string_append(command, directory->str);
        ivm_run_command(command->str, NULL, FALSE);
        g_string_free(command, TRUE);

        // wait 5 seconds for the directory to be created
        int i = 0;

        dir = opendir(directory->str);
        while (dir == NULL && i < 5)
        {
            sleep(1);
            dir = opendir(directory->str);
            i++;
        }

        if ((void *) opendir(directory->str) == NULL)
        {
            DEBUG(_("Couldn't create new directory %s\n"), directory->str);
            exit(1);
        }
        else
            closedir(dir);
    }

    else
        closedir(dir);

    g_string_free(directory, TRUE);

    char *filenames[] = {
        "IvmConfigActions.xml", "IvmConfigBase.xml",
        "IvmConfigProperties.xml",
        "IvmConfigConditions.xml", NULL
    };

    char *contents[] = {
        IVMCONFIGACTIONS, IVMCONFIGBASE,
        IVMCONFIGPROPERTIES, IVMCONFIGCONDITIONS, NULL
    };

    for (int i = 0; filenames[i] != NULL; i++)
    {
        filename = g_string_new((char *) strdup(homeDir));
        filename = g_string_append(filename, "/.ivman/");
        filename = g_string_append(filename, filenames[i]);

        file = fopen(filename->str, "r");

        if (file == NULL)
        {
            DEBUG(_
                  ("Configuration file %s not found, creating one "
                   "with default content...\n"), filename->str);

            file = fopen(filename->str, "w");

            content = contents[i];
            fwrite(content, 1, strlen(content), file);
            fclose(file);
        }
        else
            fclose(file);

        g_string_free(filename, TRUE);

    }
}



/**
 * Load IvmConfigBase.xml configuration.
 */
void ivm_load_config(void)
{
    char *file = ivm_get_config_file("IvmConfigBase.xml");

    cfg_base = parseIvmConfigBase(file);
    free(file);

    if (!cfg_base)
    {
        DEBUG(_("Could not parse IvmConfigBase.xml! "
                "Please check this file for errors! "
                "Since I need IvmConfigBase.xml to function, I "
                "will now exit.\n"));
        exit(1);
    }
    // if we're not running as root, set pidFile to ~/.ivman/.lock
    if (usermode)
    {
        GString *pidFile = g_string_new((char *) strdup(homeDir));

        pidFile = g_string_append(pidFile, "/.ivman/.lock");
        cfg_base->pidFile = (char *) strdup(pidFile->str);
        g_string_free(pidFile, TRUE);
    }
}



/**
 * Free an array of strings.
 */
void ivm_free_str_array(char **strings)
{
    if (strings == NULL)
        return;

    int i = 0;

    while (strings[i] != NULL)
    {
        free(strings[i]);
        i++;
    }

    free(strings);
}

/**
 * Run an array of commands.
 */
void ivm_run_commands(char **commands, char const *udi,
                      gboolean const execun)
{
    if (commands == NULL)
        return;

    int i = 0;

    while (commands[i] != NULL)
    {
        ivm_run_command(commands[i], udi, execun);
        i++;
    }

    if (execun)
        g_hash_table_remove(pss, udi);
}


/**
 * Run a command.
 */
gboolean ivm_run_command(char const *const command,
                         char const *udi, gboolean const execun)
{
    char *argv[4];
    char *new_command = (char *) strdup(command);

    LibHalPropertySet *hal_ps = NULL;
    DBusError dbus_error;

    dbus_error_init(&dbus_error);
    GError *error = NULL;
    GString *currentCommand;
    int length;
    int i;

    char c;
    char *substitution;
    char *propertyname = NULL;
    char *substring = strstr(new_command, "$hal.");
    char *substring2;
    int pos = (int) ((substring - new_command) / sizeof(char));
    char *beginning = (char *) strndup(new_command, pos);

    if (execun)
        hal_ps = (LibHalPropertySet *) g_hash_table_lookup(pss,
                                                           (gpointer) udi);

    while (substring != NULL)
    {
        // get rid of $hal.
        substring[0] = ' ';
        substring[1] = ' ';
        substring[2] = ' ';
        substring[3] = ' ';
        substring[4] = ' ';

        substring = (char *) g_strchug(substring);
        substring2 = strstr(substring, "$");

        length = (int) ((substring2 - substring) / sizeof(char));
        propertyname = (char *) strndup(substring, length);

        i = 0;

        while (i < length)
        {
            substring[i] = ' ';
            i++;
        }

        substring[i] = ' ';

        // hack so we don' t eat an extra space
        c = substring[i + 1];
        substring[i + 1] = 'a';

        substring = (char *) g_strchug(substring);
        substring[0] = c;



        if ((execun && psi_property_exists(hal_ps, propertyname))
            || libhal_device_property_exists(hal_ctx, udi,
                                             propertyname, &dbus_error))
        {
            unsigned int type = execun ? psi_get_property_type(hal_ps,
                                                               propertyname)
                : libhal_device_get_property_type(hal_ctx,
                                                  udi,
                                                  propertyname,
                                                  &dbus_error);

            if (type == DBUS_TYPE_STRING)
            {
                substitution =
                    execun ? (char *)
                    psi_get_property_string(hal_ps,
                                            propertyname)
                    : (char *)
                    libhal_device_get_property_string
                    (hal_ctx, udi, propertyname, &dbus_error);
            }
            else if (type == DBUS_TYPE_BOOLEAN)
            {
                if ((execun
                     && psi_get_property_bool(hal_ps,
                                              propertyname))
                    ||
                    libhal_device_get_property_bool
                    (hal_ctx, udi, propertyname, &dbus_error))
                    substitution = (char *) strdup("true");
                else
                    substitution = (char *) strdup("false");
            }
            else if (type == DBUS_TYPE_INT32 || type == DBUS_TYPE_INT64)
            {
                asprintf(&substitution, "%d",
                         execun ?
                         psi_get_property_int(hal_ps,
                                              propertyname)
                         :
                         libhal_device_get_property_int
                         (hal_ctx, udi, propertyname, &dbus_error));
            }
            else if (type == DBUS_TYPE_UINT32 || type == DBUS_TYPE_UINT64)
            {
                asprintf(&substitution, "%u",
                         execun ? (unsigned int)
                         psi_get_property_uint64(hal_ps,
                                                 propertyname)
                         : (unsigned int)
                         libhal_device_get_property_uint64
                         (hal_ctx, udi, propertyname, &dbus_error));
            }
            else
            {
                // we don 't worry about other types of
                // properties,
                // they' re pretty uncommon}
                substitution = (char *) strdup("NULL");
                DEBUG(_
                      ("Warning: unhandled HAL type encountered, NULL "
                       "substituted for value!\n"));
            }
        }
        else
        {
            DEBUG(_
                  ("Warning: nonexistent HAL property encountered, NULL "
                   "substituted for value!\n"));
            substitution = (char *) strdup("NULL");
        }



        if (propertyname != NULL)
            free(propertyname);

		// Since this will be passed as a shell argument, replace ' and " so
		// we can put the property within quotes.
		while (strstr(substitution, "'") != 0)
		{
			char * substr = strstr(substitution, "'");
			substr[0] = '?';
		}
		while (strstr(substitution, "\"") != 0)
		{
			char * substr = strstr(substitution, "\"");
			substr[0] = '?';
		}
		

        currentCommand = g_string_new(substring);
        currentCommand = g_string_prepend(currentCommand, substitution);
        currentCommand = g_string_prepend(currentCommand, beginning);
        new_command = (char *) strdup(currentCommand->str);
        g_string_free(currentCommand, TRUE);

        if (NULL != substitution)
            free(substitution);

        libhal_free_string(beginning);

        substring = strstr(new_command, "$hal.");
        int pos = (int) ((substring - new_command) / sizeof(char));

        beginning = (char *) strndup(new_command, pos);

    }

    ivm_check_dbus_error(&dbus_error);

    argv[0] = "/bin/sh";
    argv[1] = "-c";
    argv[2] = new_command;
    argv[3] = NULL;

    DEBUG(_("Running: %s\n"), argv[2]);

    if (!g_spawn_async(NULL, argv, NULL, 0, NULL, NULL, NULL, &error))
    {
        DEBUG(_("Execution of %s failed with error: %s\n"),
              new_command, error->message);
        return FALSE;
    }

    g_free(new_command);
    return TRUE;
}







// returns TRUE if DVD 'udi' is a video DVD, FALSE otherwise
gboolean ivm_is_dvd(char const *const udi)
{
    char *path;
    gboolean retval;

    DBusError dbus_error;

    dbus_error_init(&dbus_error);

    // return false if DVD is not mounted
    if (!libhal_device_property_exists
        (hal_ctx, udi, "volume.is_mounted", &dbus_error))
        return FALSE;

    if (libhal_device_get_property_bool
        (hal_ctx, udi, "volume.is_mounted", &dbus_error) == FALSE)
        return FALSE;

    if (!libhal_device_property_exists
        (hal_ctx, udi, "volume.mount_point", &dbus_error))
        return FALSE;

    char *mount_point = (char *)
        libhal_device_get_property_string(hal_ctx, udi,
                                          "volume.mount_point",
                                          &dbus_error);

    if (!libhal_device_property_exists
        (hal_ctx, udi, "block.device", &dbus_error))
        return FALSE;

    char *device = libhal_device_get_property_string(hal_ctx,
                                                     udi,
                                                     "block.device",
                                                     &dbus_error);

    DEBUG(_("Checking for video DVD in device '%s' mounted at '%s'\n"),
          device, mount_point);

    path = g_build_path(G_DIR_SEPARATOR_S, mount_point, "video_ts", NULL);
    retval = g_file_test(path, G_FILE_TEST_IS_DIR);
    g_free(path);

    /* try the other name, if needed */

    if (retval == FALSE)
    {
        path =
            g_build_path(G_DIR_SEPARATOR_S, mount_point, "VIDEO_TS", NULL);
        retval = g_file_test(path, G_FILE_TEST_IS_DIR);
        g_free(path);
    }

    if (retval == TRUE)
        DEBUG(_("%s looks like a video DVD\n"), device);
    else
        DEBUG(_("%s does not look like a video DVD\n"), device);

    libhal_free_string(mount_point);
    libhal_free_string(device);
    ivm_check_dbus_error(&dbus_error);

    return retval;
}




/**
 * Mount device 'udi'
 */
void ivm_device_mount(char const *const udi)
{
    if (mount_command == NULL)
    {
        DEBUG(_("Can't mount %s; no mount command available!\n"), udi);
        return;
    }

    ivm_run_command(mount_command, udi, FALSE);
    return;
}



/**
 * Handle media change / device attach on 'udi'
 */
void ivm_media_changed(char const *const udi)
{
    DBusError dbus_error;

    dbus_error_init(&dbus_error);

    char *file = ivm_get_config_file("IvmConfigActions.xml");
    IvmConfigActions *cfg = parseIvmConfigActions(file, (char *) udi);

    free(file);

    if (!cfg)
    {
        xmlReadFailed("IvmConfigActions.xml");
        return;
    }

    char *device = NULL;

    if (libhal_device_property_exists
        (hal_ctx, udi, "block.device", &dbus_error))
        device =
            (char *)
            libhal_device_get_property_string(hal_ctx,
                                              udi,
                                              "block.device", &dbus_error);

    if (cfg->mount == TRUE)
    {
        // If we 're running in system-wide mode, sleep a few
        // seconds
        // to allow any user-mode instances to mount first.

        if (!usermode && cfg_base->sleep)
        {
            DEBUG(_("Giving other programs a chance to mount...\n"));
            sleep(5);
        }

        if (libhal_device_property_exists(hal_ctx, udi,
                                          "volume.is_mounted",
                                          &dbus_error)
            && !libhal_device_get_property_bool(hal_ctx, udi,
                                                "volume.is_mounted",
                                                &dbus_error))
        {
            DEBUG(_("Attempting to mount device %s\n"), device);
            ivm_device_mount(udi);
            sleep(1);
        }
    }

    if ((!libhal_device_property_exists(hal_ctx, udi,
                                        "volume.is_mounted",
                                        &dbus_error))
        ||
        (!libhal_device_get_property_bool
         (hal_ctx, udi, "volume.is_mounted", &dbus_error))
		||
		(!libhal_device_property_exists(hal_ctx, udi, "volume.mount_point", &dbus_error))
	   )
    {
        DEBUG(_("%s wasn't mounted, by us or by others..."), udi);
    }
    else 
    {
        // After mounting the filesystem, it takes HAL a while to
        // update, so we poll once every second for five seconds
        // to find out the mount point

        int i = 0;

        char *mount_point = (char *)
            libhal_device_get_property_string(hal_ctx, udi,
                                              "volume.mount_point",
                                              &dbus_error);

        while (((mount_point == NULL) || (strlen(mount_point) < 1))
               && (i < 5))
        {
            sleep(1);
            mount_point = (char *)
                libhal_device_get_property_string(hal_ctx, udi,
                                                  " volume.mount_point ",
                                                  &dbus_error);
            i++;
        }


        if (mount_point == NULL || strlen(mount_point) < 1)
        {
            DEBUG(_("Couldn't get mount point of device %s after "
                    "5 seconds\n"), device);
        }

        libhal_free_string(mount_point);
    }

    if (TRUE == ivm_is_dvd(udi))
        ivm_run_commands(cfg->execdvd, udi, FALSE);
    ivm_free_str_array(cfg->execdvd);

    ivm_run_commands(cfg->exec, udi, FALSE);
    ivm_free_str_array(cfg->exec);

    if (cfg->execun != NULL)
    {
        g_hash_table_insert(execuns, (gpointer)
                            strdup(udi), (gpointer) cfg->execun);
    }
    LibHalPropertySet *hal_ps =
        libhal_device_get_all_properties(hal_ctx, udi, &dbus_error);

    g_hash_table_insert(pss, (gpointer) strdup(udi), (gpointer) hal_ps);

    free(cfg);

    libhal_free_string(device);
    ivm_check_dbus_error(&dbus_error);
}



/** Unmount if required */
void ivm_umount_if_needed(char const *const udi)
{
    if (udi == NULL)
        return;

    DEBUG(_("Unmounting %s\n"), udi);

    ivm_run_command(umount_command, udi, TRUE);
}

/**
  * Set up some gettext stuff.
  */
void setupI18N()
{
    if (setlocale(LC_MESSAGES, "") == NULL)
        DEBUG(_("setlocale() failed!\n"));

    if (bindtextdomain(PACKAGE, DATADIR) == NULL)
        DEBUG(_("bindtextdomain() failed!\n"));

    if (textdomain(PACKAGE) == NULL)
        DEBUG(_("textdomain() failed!\n"));
}


/**
 * Set up connection to HAL.
 */
void setupHAL()
{
    DBusError dbus_error;

    dbus_error_init(&dbus_error);

#ifdef HAL_0_4
    LibHalFunctions hal_functions = {
        hal_mainloop_integration,
        hal_device_added,
        hal_device_removed,
        hal_device_new_capability,
        hal_device_lost_capability,
        hal_device_property_modified, hal_device_condition
    };

    hal_ctx = (LibHalContext *) hal_initialize(&hal_functions, FALSE);
    if (!hal_ctx)
    {
        DEBUG(_("Failed to create HAL context!\n"));
        exit(1);
    }
#else
    hal_ctx = (LibHalContext *) libhal_ctx_new();
    if (!hal_ctx)
    {
        DEBUG(_("Failed to create HAL context!\n"));
        exit(1);
    }

    if (!hal_mainloop_integration(hal_ctx, &dbus_error))
    {
        DEBUG(_("Couldn't connect to HAL!\n"));
        exit(1);
    }

    // Set up all the callbacks in hal_interface.c
    libhal_ctx_set_device_added(hal_ctx, hal_device_added);
    libhal_ctx_set_device_removed(hal_ctx, hal_device_removed);
    libhal_ctx_set_device_new_capability(hal_ctx,
                                         hal_device_new_capability);
    libhal_ctx_set_device_lost_capability(hal_ctx,
                                          hal_device_lost_capability);
    libhal_ctx_set_device_property_modified(hal_ctx,
                                            hal_device_property_modified);
    libhal_ctx_set_device_condition(hal_ctx, hal_device_condition);

    if (!libhal_ctx_init(hal_ctx, &dbus_error))
    {
        DEBUG(_("Couldn't initialise HAL!\n"));
        exit(1);
    }
#endif



#ifdef HAL_0_4
    if (libhal_device_property_watch_all(hal_ctx, &dbus_error))
#else
    if (!libhal_device_property_watch_all(hal_ctx, &dbus_error))
#endif                          // HAL_0_4
    {
        DEBUG(_("Failed to watch all HAL properties!\n"
                "Error: %s\n"), dbus_error.message);
        exit(1);
    }

    devices = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
    execuns = g_hash_table_new_full(g_str_hash, g_str_equal,
                                    free,
                                    (GDestroyNotify) ivm_free_str_array);
#ifndef HAL_0_4

    pss =
        g_hash_table_new_full(g_str_hash,
                              g_str_equal, free,
                              (GDestroyNotify) libhal_free_property_set);
#else

    pss =
        g_hash_table_new_full(g_str_hash,
                              g_str_equal, free,
                              (GDestroyNotify) hal_free_property_set);
#endif

    // Now try to watch all properties, if necessary.
    char *file = ivm_get_config_file("IvmConfigProperties.xml");
    IvmConfigProperties *prop_cfg =
        parseIvmConfigProperties(file, NULL, NULL);

    if (!prop_cfg)
    {
        xmlReadFailed(" IvmConfigProperties.xml ");
    }

    else
    {
        if (prop_cfg->checkOnInit == TRUE)
        {
            DEBUG(_
                  (" Running through rules for every current property.\n"));

            // run through every device
            int num_devices;
            char **devices =
                (char **) libhal_get_all_devices(hal_ctx, &num_devices,
                                                 &dbus_error);

            int i = 0;

            while (i < num_devices)
            {
				//DEBUG(_("Currently checking device %s"), devices[i]);
                IvmConfigProperties *newProp_cfg =
                    IvmConfigPropertiesAll(file, devices[i]);

                ivm_run_commands(newProp_cfg->exec, devices[i], FALSE);
                ivm_free_str_array(newProp_cfg->exec);
                free(newProp_cfg);

                i++;
            }

            if (devices != NULL)
                libhal_free_string_array(devices);

        }


        free(prop_cfg);
    }

    free(file);

    ivm_check_dbus_error(&dbus_error);
}



/**
 * Parse command-line options.
 */
void parseArguments(int argc, char *argv[])
{
    GError *optParseError = NULL;
    char *systemHelpText;

    asprintf(&systemHelpText,
             _("Force Ivman to run in system-wide mode. Use this if you "
               "want Ivman to read configuration from %s/ivman/ but you "
               "are not running it as root."), SYSCONFDIR);
    GOptionEntry entries[] = {
        {"system", 's', 0,
         G_OPTION_ARG_NONE,
         &systemOpt,
         systemHelpText, NULL},
        {"debug", 'd', 0,
         G_OPTION_ARG_NONE,
         &debugOpt,
         _("Force Ivman to show debugging output."),
         NULL},
        {"nofork", (char) NULL, 0,
         G_OPTION_ARG_NONE,
         &noForkOpt,
         _("Force Ivman not to daemonize itself."),
         NULL},
        {(const gchar *) NULL,
         (gchar) NULL,
         (gint) NULL,
         (GOptionArg) NULL,
         (gpointer) NULL, (const gchar *) NULL,
         (const gchar *) NULL}
    };

    GOptionContext *context = g_option_context_new(_("- start ivman"));

    g_option_context_add_main_entries(context, entries, NULL);

    if (!g_option_context_parse(context, &argc, &argv, &optParseError))
    {
        DEBUG("%s\n", optParseError->message);
        exit(1);
    }

    free(systemHelpText);
}










/**
 * Do all configuration necessary for initialisation.
 */
void do_startup_configure()
{
    // Assume user mode
    usermode = TRUE;

    // are we running as root or with --system option?
    if (systemOpt || geteuid() == 0)
        usermode = FALSE;

    homeDir = (gchar *) g_get_home_dir();

    // Create default configs if necessary
    if (usermode)
        ivm_test_configs();

    // Load configuration
    ivm_load_config();

    if (debugOpt)
        cfg_base->debug = TRUE;

    isdaemon = FALSE;

    signal(SIGHUP, SIG_IGN);

    if (noForkOpt)
        cfg_base->fork = FALSE;



    // Find mount command to use.
    set_mount_command();
    if ((mount_command == NULL) || (umount_command == NULL))
    {
        DEBUG(_("An appropriate mount command could not be found! "
                "Please make sure you have mount/umount or pmount/pumount in "
                "Ivman's PATH. Mounting will not work.\n"));
    }

    // Permanently unlock CD / DVD drive
    if (geteuid() == 0)
        ivm_run_command("echo 0 > /proc/sys/dev/cdrom/lock", NULL, FALSE);


}



/**
 * main function.
 */
int main(int argc, char *argv[])
{
    setupI18N();
    parseArguments(argc, argv);
    do_startup_configure();

    // Connect to HAL
    setupHAL();

    DEBUG("%s, http:/ivman.sourceforge.net\n", PACKAGE_STRING);

#ifdef HAL_0_4
    DEBUG(_("Compiled against HAL 0.4.x or earlier\n"));
#else
    DEBUG(_("Compiled against HAL 0.5.x or later\n"));
#endif

    if (usermode == FALSE)
    {
        DEBUG(_("Running in system mode.\n"));
    }
    else
    {
        DEBUG(_("Running in user mode.\n"));
    }


    if (cfg_base->fork == TRUE)
    {
        // Daemonize
        isdaemon = daemonize();

        if (!isdaemon)
        {
			DEBUG(_("Couldn't daemonize, exiting..."));
            return 1;
        }
    }

    // Drop privileges.  Must do this _after_ daemonizing!
    if ((geteuid() == 0) && (!dropPrivileges(cfg_base->user, cfg_base->group)))
    {
        DEBUG(_("Couldn't drop privileges to %s:%s, exiting!"),
              cfg_base->user, cfg_base->group);
        return 1;
    }
	
    GMainLoop *loop = g_main_loop_new(NULL, FALSE);

    DEBUG(_("Entering main loop.\n"));
    g_main_loop_run(loop);

    // When we get here, main loop is finished.
    if (cfg_base->fork == TRUE)
        clear_pidfile(cfg_base->pidFile);

    g_hash_table_destroy(devices);

    return 0;

}
